@@ -19,7 +19,7 @@ class SmartFormatter(argparse.HelpFormatter):
 
 
 class BlankTrue(argparse.Action):
-    """ When left blank, this flag produces True. (Normal behaviour is to produce None which I use for not being set."""
+    """ When left blank, this flag produces True. Normal behaviour is to produce None which I use for not being set."""
 
     def __call__(self, _, namespace, values, option_string=None):
         if values in [None, []]:  # blank argument with nargs="?" produces None, with ="*" produces []
@@ -27,6 +27,28 @@ class BlankTrue(argparse.Action):
         setattr(namespace, self.dest, values)
 
 
+class BlankTrueFalseStr(argparse.Action):
+    """ When left blank, this flag produces True. Normal behaviour is to produce None which I use for not being set.
+    When 0/false/off 1/true/on used, bool is produced, else str is taken.
+    """
+
+    def __call__(self, _, namespace, values, option_string=None):
+        allow_string = True
+        if values in [None, []]:  # blank argument with nargs="?" produces None, with ="*" produces []
+            values = True
+        elif values.lower() in ["0", "false", "off"]:
+            values = False
+        elif values.lower() in ["1", "true", "on"]:
+            values = True
+        elif not allow_string \
+                and (type(self.metavar) is not list or values.lower() not in self.metavar) \
+                and (len(self.metavar.split("/")) < 2 or values.lower() not in self.metavar.split("/")):
+            print(f"Unrecognised value '{values}' of '{self.dest}'. Allowed values are 0/1/BLANK."
+                  f" Should the value be considered a positional parameter, move '{self.dest}' behind.")
+            exit()
+        setattr(namespace, self.dest, values)
+
+
 def _get_envelope(instance: Envelope, args):
     """ Internal method. If loaded from STDIN, Envelope object exists already otherwise new object is created.  """
     # noinspection PyProtectedMember
@@ -36,7 +58,8 @@ def _get_envelope(instance: Envelope, args):
 def main():
     parser = argparse.ArgumentParser(description=__doc__, formatter_class=SmartFormatter)
     group_io = parser.add_argument_group("Input/Output")
-    group_io.add_argument('--message', help='Plain text message.', metavar="TEXT", nargs="?", action=BlankTrue)
+    group_io.add_argument('--message', help='Plain text message. Empty to read.',
+                          metavar="TEXT", nargs="?", action=BlankTrue)
     group_io.add_argument('--input', help='Path to message file. (Alternative to `message` parameter.)', metavar="FILE")
     group_io.add_argument('--output',
                           help='Path to file to be written to (else the contents is returned if ciphering or True if sending).',
@@ -76,21 +99,26 @@ def main():
     group_ciph.add_argument('--attach-key', help="Appending public key to the attachments when sending.",
                             action="store_true")
 
-    group_send = parser.add_argument_group("Sending")
-    group_send.add_argument('-s', '--subject', help="E-mail subject", nargs="?", action=BlankTrue)
-    group_send.add_argument('-t', '--to', help="E-mail – needed to choose their key if encrypting", metavar="E-MAIL",
+    group_recip = parser.add_argument_group("Recipients")
+    group_recip.add_argument('-t', '--to', help="E-mail – needed to choose their key if encrypting", metavar="E-MAIL",
                             nargs="*", action=BlankTrue)
-    group_send.add_argument('--cc', help="E-mail or list", metavar="E-MAIL", nargs="*", action=BlankTrue)
-    group_send.add_argument('--bcc', help="E-mail or list", metavar="E-MAIL", nargs="*", action=BlankTrue)
-    group_send.add_argument('--reply-to',
+    group_recip.add_argument('--cc', help="E-mail or list", metavar="E-MAIL", nargs="*", action=BlankTrue)
+    group_recip.add_argument('--bcc', help="E-mail or list", metavar="E-MAIL", nargs="*", action=BlankTrue)
+    group_recip.add_argument('--reply-to',
                             help="Header that states e-mail to be replied to. The field is not encrypted.",
                             metavar="E-MAIL", nargs="?", action=BlankTrue)
-    group_send.add_argument('-f', '--from', help="E-mail – needed to choose our key if encrypting", metavar="E-MAIL",
+    group_recip.add_argument('-f', '--from', help="E-mail – needed to choose our key if encrypting", metavar="E-MAIL",
                             nargs="?", action=BlankTrue)
-    group_send.add_argument('--sender', help="Alias for --from if not set."
-                                             " Otherwise appends the \"Sender\" header.", metavar="E-MAIL")
-    group_send.add_argument('--no-sender', action="store_true",
+    group_recip.add_argument('--sender', help="Alias for --from if not set."
+                                              " Otherwise appends the \"Sender\" header.", metavar="E-MAIL")
+    group_recip.add_argument('--no-sender', action="store_true",
                             help="We explicitly say we do not want to decipher later if encrypting.")
+
+    group_send = parser.add_argument_group("Sending")
+    group_send.add_argument('-s', '--subject', help="E-mail subject", nargs="?", action=BlankTrue)
+    group_send.add_argument('--subject-encrypted', help="Text used instead of the real protected subject"
+                                                        " while PGP encrypting. Put 0/false/off to not encrypt.",
+                            action=BlankTrueFalseStr)
     group_send.add_argument('-a', '--attach',
                             help="Path to the attachment, followed by an optional file name to be used and/or mimetype."
                                  " This parameter may be used multiple times.",
@@ -234,6 +262,7 @@ def main():
         # XX allow any header to be displayed, ex: `--header Received` will display all Received headers
         read_method = None
         read_val = None
+        # if some of the following keys are true, we want to read that value instead of setting it
         for x in (x for x in ("subject", "message", "from_", "to", "cc", "bcc", "reply_to") if args[x] is True):
             read_method = x
             del args[x]
