@@ -18,7 +18,7 @@ from email.utils import make_msgid, formatdate, getaddresses
 from getpass import getpass
 from itertools import chain
 from pathlib import Path
-from quopri import decodestring
+from quopri import decodestring, decode as quopri_decode
 from typing import Union, List, Set, Any
 
 from .utils import Address, Attachment, AutoSubmittedHeader, SMTP, _Message, is_gpg_fingerprint, assure_list, \
@@ -72,9 +72,7 @@ class Envelope:
         return self._status
 
     def __str__(self):
-        if self._result_cache_hash and self._result_cache_hash != self._param_hash():
-            # ex: if we change Subject, we have to regenerate self._result
-            self._result.clear()
+        self._result_fresh()
         if not self._result:
             if self._encrypt or self._sign:
                 # if subject is not set, we suppose this is just a data blob to be encrypted, no an e-mail message
@@ -90,12 +88,12 @@ class Envelope:
     def __repr__(self):
         """
         :return: Prints out basic representation.
-            However this is not serialization, you cannot reconstruct any complicated objects having attachments or custom headers.
+            However this is not a serialization: you cannot reconstruct any complicated objects having attachments or custom headers.
         """
         l = []
         quote = lambda x: '"' + x.replace('"', r'\"') + '"' if type(x) is str else x
 
-        text, html = self._message.get()
+        text, html = self._message.get(str)
         message = {}
         if text and html:
             message = {"message(html)": html,
@@ -116,13 +114,11 @@ class Envelope:
         return f"Envelope({', '.join(l)})"
 
     def __bytes__(self):
-        if not self._result:
-            str(self)
+        self._result_fresh(True)
         return assure_fetched(self._get_result_str(), bytes)
 
     def __eq__(self, other):
-        if not self._result:
-            str(self)
+        self._result_fresh(True)
         me = assure_fetched(self._get_result_str(), bytes)
         if type(other) in [str, bytes]:
             return me == assure_fetched(other, bytes)
@@ -134,8 +130,7 @@ class Envelope:
             Bcc and attachments are mentioned.
             Ex: whilst we have to use quoted-printable, here the output will be plain text.
         """
-        if not self._result:
-            str(self)
+        self._result_fresh(True)
         result = []
         for a in self._attachments:  # include attachments info as they are removed with the payload later
             if a.inline:
@@ -159,7 +154,7 @@ class Envelope:
             else:
                 result.append(r)
 
-        text, html = self._message.get()
+        text, html = self._message.get(str)
         if text and html:
             result.extend(["* MESSAGE VARIANT text/plain:", text, "",
                            "* MESSAGE VARIANT text/html:", html])
@@ -174,10 +169,18 @@ class Envelope:
             self._result_cache = s  # slightly quicker next time if ever containing a huge amount of lines
         return self._result_cache
 
+    def _result_fresh(self, recreate=False):
+        if self._result_cache_hash and self._result_cache_hash != self._param_hash():
+            # ex: if we change Subject, we have to regenerate self._result
+            self._result.clear()
+        if recreate and not self._result:
+            str(self)
+
     def as_message(self) -> Message:
         """
         :return: Message object is S/MIME is used, EmailMessage otherwise.
         """
+        self._result_fresh()
         for el in self._result:
             if isinstance(el, Message):
                 return el
@@ -209,7 +212,6 @@ class Envelope:
             return Parser(e, key=key, cert=cert, gnupg_home=e._get_gnupg_home()).parse(o, add_headers=True)
         except ValueError as err:
             logger.warning(f"Message might not have been loaded correctly. {err}")
-            import ipdb; ipdb.post_mortem()
 
         # emergency body loading when parsing failed
         header_row = re.compile(r"([^\t:]+):(.*)")
@@ -245,7 +247,7 @@ class Envelope:
 
     def __init__(self, message=None, from_=None, to=None, subject=None, headers=None,
                  gpg=None, smime=None,
-                 encrypt=None, sign=None, passphrase=None, attach_key=None, cert=None,
+                 encrypt=None, sign=None, passphrase=None, attach_key=None, cert=None, subject_encrypted=None,
                  sender=None, cc=None, bcc=None, reply_to=None, mime=None, attachments=None,
                  smtp=None, output=None, send=None):
         """
@@ -269,6 +271,7 @@ class Envelope:
 
         Encrypting
         :param encrypt: Recipients public key string or Path or stream (ex: from open()).
+        :param subject_encrypted: Text used instead of the real protected subject while PGP encrypting. False to not encrypt.
         :param to: E-mail or list. If encrypting used so that we choose the key they will be able to decipher with.
         :param from_: E-mail of the sender. If encrypting used so that we choose our key to be still able
                         to decipher the message later with.
@@ -301,6 +304,9 @@ class Envelope:
         #   GPG: (True, key contents, fingerprint, None)
         #   SMIME: certificate contents
         self._encrypt = None
+
+        # `_from` and `_sender` properties might be false because of `sender=False` attribute (or "--no-sender" flag)
+        # that explicitly states we have no sender
         self._from: Union[Address, False] = None
         self.__from: Union[Address, False] = None
         self._sender: Union[Address, False] = None
@@ -310,6 +316,7 @@ class Envelope:
         self._bcc: List[Address] = []
         self._reply_to: List[Address] = []
         self._subject: str = None
+        self._subject_encrypted: Union[str, False] = True
         self._smtp = None
         self._attachments: List[Attachment] = []
         self._mime = AUTO
@@ -367,6 +374,8 @@ class Envelope:
                 self.signature(v)
             elif k == "encrypt":
                 self.encryption(v)
+            elif k == "subject_encrypted":
+                self.subject(encrypted=v)
             elif k == "_Message":  # internal stuff
                 continue
             elif v is not None and v != []:  # "to" will receive [] by default
@@ -428,7 +437,7 @@ class Envelope:
         """ An alias of .message """
         return self.message(text=text, path=path)
 
-    def message(self, text=None, *, path=None, alternative=AUTO, boundary=None) -> Union["Envelope", Any]:
+    def message(self, text=None, *, path=None, alternative=AUTO, boundary=None) -> Union["Envelope", str]:
         """
         Message to be ciphered / e-mail body text.
         :param text: Any attainable contents.
@@ -453,8 +462,8 @@ class Envelope:
             # Content-Type: text/html; charset="utf-8"
             # He<b>llo</b>
 
-        :return: Envelope if text or path is set
-            else return anything that has been inserted to the .message() before (probably str, bytes)
+        :return: Envelope if `text` or `path` is set
+            otherwise return anything that has been inserted to the .message() before as `str`
         """
         # XX make preview default over send(0) when no action is given?
         if boundary is not None:
@@ -463,30 +472,41 @@ class Envelope:
             raise ValueError(f"Invalid alternative {alternative} for message, choose one of the: {AUTO}, {PLAIN}, {HTML}")
         if text is path is None:
             # reading value
-            s = getattr(self._message, alternative)
-            if not s and alternative == AUTO:  # prefer reading HTML over plain text if alternative set
-                s = self._message.html or self._message.plain
+            b = getattr(self._message, alternative)
+            if not b and alternative == AUTO:  # prefer reading HTML over plain text if alternative set
+                b = self._message.html or self._message.plain
 
+            # When loading an EML, Content-Type and Content-Transfer-Encoding get wiped off and the message is decoded.
+            # However, the user can define the headers explicitly.
+            # In that case we expect the message is already inserted encoded. (However, if it not, we do our best.)
+
+            # The user can define the Content-Type
+            content_type = self._headers.get("Content-Type")
+            m = re.search("charset=(.*)", content_type, re.IGNORECASE) if content_type else None
+            charset = m[1] if m else "utf-8"
+
+            # Content-Transfer-Encoding might be active
             transfer = self._headers.get("Content-Transfer-Encoding")
-            # Useful for reading loaded EML that might have been encoded.
-            # When loading an EML, everything was loaded into `self._message.auto`.
-            # XX When loading an EML, everything is under .auto so that quering alternative="html"|"plain" would return None here!
             if transfer == "base64":
                 try:
-                    return b64decode(s).decode("utf-8")
+                    return b64decode(b).decode(charset)
                 except binascii.Error:
-                    pass
+                    pass  # the user specified Content-Transfer-Encoding but left the message unencoded
                 except UnicodeDecodeError:
-                    raise TypeError(f"Cannot base64-decode the message: {s}")
+                    raise TypeError(f"Cannot base64-decode the message: {b}")
             elif transfer == "quoted-printable":
                 try:
-                    return decodestring(s).decode("utf-8")
+                    return decodestring(b).decode(charset)
                 except ValueError:
-                    pass
+                    pass  # the user specified Content-Transfer-Encoding but left the message unencoded
+
+            if b is None:
+                return ""
+            try:
+                return b.decode(charset)
+            except UnicodeError:
+                raise ValueError(f"Cannot decode the message correctly, it is not in Unicode. {b}")
 
-            # XX this should return str or bytes
-            # We prefer string. But we want to keep it as encoding agnostic as possible so bytes might be better.
-            return s or ""  # `s` might be None
 
         # write value
         if type(text) is _Message:  # constructor internally adopts default's self.default._message
@@ -496,12 +516,9 @@ class Envelope:
         if path:
             text = Path(path)
 
-        setattr(self._message, alternative, assure_fetched(text, str))
+        setattr(self._message, alternative, assure_fetched(text, bytes))
         return self
 
-        # self._message = text
-        # return self
-
     def date(self, date):
         """
         Specify Date header. If not used, Date will be added automatically.
@@ -519,14 +536,15 @@ class Envelope:
     def sender(self, email=None) -> Union["Envelope", Address]:
         """  Alias for "from" if not set. Otherwise appends header "Sender". If None, current `Sender` returned. """
         if email is None:
-            return self.__sender
+            return self.__sender or Address()
         self._sender = Address.parse(email, single=True, allow_false=True)
         self._prepare_from()
         return self
 
     def from_(self, email=None) -> Union["Envelope", Address]:
+        """ Set the sender header. If None, current `From` returned. """
         if email is None:
-            return self.__from
+            return self.__from or Address()
         self._from = Address.parse(email, single=True, allow_false=True)
         self._prepare_from()
         return self
@@ -546,10 +564,18 @@ class Envelope:
         self._gpg = False
         return self
 
-    def subject(self, subject=None) -> Union["Envelope", str]:
-        if subject is None:
+    def subject(self, subject=None, encrypted: Union[str, bool] = None) -> Union["Envelope", str]:
+        """ Get or set the message subject
+        :param subject: Subject text.
+        :param encrypted: Text used instead of the real protected subject while PGP encrypting. False to not encrypt.
+        :return If neither parameter specified, current subject returned, otherwise return self.
+        """
+        if subject is None and encrypted is None:
             return str(self._subject or "")
-        self._subject: str = subject
+        if subject is not None:
+            self._subject: str = subject
+        if encrypted is not None:
+            self._subject_encrypted = encrypted
         return self
 
     def mime(self, subtype=AUTO, nl2br=AUTO):
@@ -612,14 +638,14 @@ class Envelope:
 
     auto_submitted: AutoSubmittedHeader
 
-    def header(self, key, val=None, replace=False):
+    def header(self, key, val=None, replace=False) -> Union["Envelope", list, str, None]:
         """ Add a generic header.
         The header will not be encrypted with GPG nor S/MIME.
         :param key: str Header name
-        :param val: str Header value. If None, currently used value is returned.
+        :param val: str Header value. If None, currently used value is returned as string, or their list, or None.
         :param replace: bool If True, any header of the `key` name are removed first and if `val` is None, the header is deleted.
                         Otherwise another header of the same name is appended.
-        :return: Envelope|str|list Returned self if `val` is not None or replace=True, else returns value of the header
+        :return: Envelope|str|list|None Returned self if `val` is not None or replace=True, else returns value of the header
                  or its list if the header was used multiple times. (Note that cc and bcc headers always return list.)
         """
 
@@ -662,7 +688,11 @@ class Envelope:
                 return h[0]
             return h
         elif val:
-            self._headers[key] = val
+            try:
+                self._headers[key] = val
+            except TypeError as e:
+                logger.warning(f"Header {key} could not be successfully loaded with {val}: {e}")
+                self._headers[key] = str(val)
         return self
 
     def smtp(self, host="localhost", port=25, user=None, password=None, security=None):
@@ -878,8 +908,8 @@ class Envelope:
         """ Start processing. Either sign, encrypt or send the message and possibly set bool status of the object to True.
         * send == SIMULATION is the same as send == False but the message "have not been sent" will not be produced
         """
-        text: str
-        html: str
+        plain: bytes
+        html: bytes
         data: bytes
 
         self._status = False
@@ -901,10 +931,10 @@ class Envelope:
             return
 
         # assure streams are fetched and files are read from their paths
-        text, html = self._message.get()
+        plain, html = self._message.get()
 
         # we need a message
-        if not any((text, html)):
+        if not any((plain, html)):
             logger.error("Missing message")
             return
 
@@ -914,12 +944,12 @@ class Envelope:
         # if we plan to send later, convert text message to the email message object
         email = None
         if send is not None or html:  # `html` means the user wants a 'multipart/alternative' e-mail message
-            email = self._prepare_email(text, html, encrypt and gpg_on, sign and gpg_on, sign)
+            email = self._prepare_email(plain, html, encrypt and gpg_on, sign and gpg_on, sign)
             if not email:
                 return
             data = email.as_bytes()
         else:
-            data = text.encode("utf-8")
+            data = plain
 
         # with GPG, encrypt or sign either text message or email message object
         micalg = None
@@ -981,10 +1011,10 @@ class Envelope:
                 if sign:
                     if sign in [True, AUTO]:  # try to determine sign based on the "From" header
                         fallback_sign = sign = None
-                        try:
-                            address_searched = self.__from.address
-                        except AttributeError:
-                            # there is no "From" header and no default key is given, pick the first secret as a default
+                        address_searched = self.__from.address if self.__from else False
+                        if not address_searched:
+                            # there is no "From" header (or the "From" header address is empty)
+                            # and no default key is given, pick the first secret as a default
                             for key in self._gnupg.list_keys(True):
                                 fallback_sign = key["keyid"]
                                 break
@@ -1018,7 +1048,8 @@ class Envelope:
 
     def _send_now(self, email, encrypt, encrypted_subject, send):
         try:
-            if not self.__from and send is True:
+            if not self.__from and self.__from is not False and send is True:
+                # allow ignore sender when deliberately set to False
                 logger.error("You have to specify sender e-mail.")
                 return False
             if self.__from:
@@ -1039,9 +1070,7 @@ class Envelope:
         # insert arbitrary headers
         # XX do not we want to encrypt these headers with GPG/SMIME?
         for k, v in self._headers.items():
-            # XXX check e-mail headers are really case insensitive when loading, add tests
-            # if k in ["Content-Type", "Content-Transfer-Encoding", "MIME-Version"]:
-            if k.lower() in ["content-type", "content-transfer-encoding", "mime-version"]:
+            if k.lower() in ("content-type", "content-transfer-encoding", "mime-version"):
                 # skip headers already inserted in _prepare_email
                 continue
             try:
@@ -1077,7 +1106,8 @@ class Envelope:
 
     def _param_hash(self):
         """ Check if headers changed from last _start call."""
-        return hash(frozenset(self._headers.items())) + hash("".join(self.recipients())) + hash(self._subject) + hash(self.__from)
+        return (hash(frozenset(self._headers.items())) + hash("".join(self.recipients()))
+                + hash(self._subject) + hash(self._subject_encrypted) + hash(self.__from))
 
     def _sign_gpg_now(self, message, sign, send):
         status = self._gnupg.sign(
@@ -1233,11 +1263,12 @@ class Envelope:
         email.attach(msg_signature)
         return email
 
-    @staticmethod
-    def _compose_gpg_encrypted(text):
+    def _compose_gpg_encrypted(self, text):
         # encrypted message structure according to RFC3156
         email = EmailMessage()
-        email["Subject"] = "Encrypted message"  # real subject should be revealed when decrypted
+        # real subject might be hidden until decrypted
+        email["Subject"] = ("Encrypted message" if self._subject_encrypted is True else self._subject_encrypted)\
+            if self._subject_encrypted else self._subject
         email.set_type("multipart/encrypted")
         email.set_param("protocol", "application/pgp-encrypted")
         msg_version = EmailMessage()
@@ -1252,7 +1283,7 @@ class Envelope:
         email.attach(msg_text)
         return email
 
-    def _prepare_email(self, text: str, html: str, encrypt_gpg, sign_gpg, sign):
+    def _prepare_email(self, plain: bytes, html: bytes, encrypt_gpg, sign_gpg, sign):
         """
         :type sign: If GPG, this should be the key fingerprint.
         """
@@ -1266,11 +1297,16 @@ class Envelope:
             # determine mime subtype and maybe do nl2br
             mime, nl2br = self._mime, self._nl2br
 
-            if text is None:
-                text, html = html, None
+            if plain is None:
+                plain, html = html, None
                 mime = HTML
 
-            t: str = text
+            try:
+                t: str = plain.decode()
+            except UnicodeError:
+                raise ValueError("Cannot decode the message correctly, it is not in Unicode."
+                                 " Either specify Content-Type and Content-Transfer-Encoding headers manually"
+                                 " or pass the message as str.")
             if mime == AUTO:
                 if html:
                     mime = PLAIN
@@ -1301,10 +1337,17 @@ class Envelope:
                 msg_text["Content-Transfer-Encoding"] = self._headers["Content-Transfer-Encoding"]
             # When the user sets Content-Type as multipart, they signalize they want to create sub-message themselves,
             # if they do not do that, the message looks badly. Ex: the text should contain boundaries and sub-messages.
-            msg_text.set_payload(text, "utf-8")
-            # msg_text.add_alternative(text, "utf-8")
+            msg_text.set_payload(plain)
 
         if html:
+            # prefer transferring the HTML alternative as a string,
+            # so the source code of the EML is more readable (no Content-Transfer-Encoding is used)
+            try:
+                # make sure `html` is in Unicode
+                html_s = html.decode()
+            except UnicodeError:
+                raise ValueError("Cannot decode the text/html alternative bytes correctly, it is not in Unicode."
+                                 " Pass it as str instead.")
             try:
                 # if a line is longer than 1000 characters, force EmailMessage to encode whole message
                 if any(line for line in html.splitlines() if len(line) >= 1000):
@@ -1314,12 +1357,12 @@ class Envelope:
 
                     # create an alternative message part and set utf-8 encoding explicitly
                     alt_msg = EmailMessage()
-                    alt_msg.set_content(html.encode("utf-8"), maintype="text", subtype="html")  # `html` as bytes
+                    alt_msg.set_content(html_s.encode("utf-8"), maintype="text", subtype="html")  # pass `html` as bytes
                     alt_msg.set_param("charset", "utf-8", replace=True)
                     msg_text.make_alternative()
                     msg_text.attach(alt_msg)
                 else:
-                    msg_text.add_alternative(html, subtype='html')  # `html` as string
+                    msg_text.add_alternative(html_s, subtype='html')  # `html` as string
             except (ValueError, TypeError):
                 # Content-Type: multipart/mixed -> ValueError: Cannot convert mixed to alternative
                 # Content-Type: multipart/alternative -> TypeError: Attach is not valid on a message with a non-multipart payload
@@ -1494,11 +1537,8 @@ class Parser:
     def parse(self, o: Message, add_headers=False):
         if add_headers:
             for k, val in o.items():
-                # XXX
-                # print(k, val)
-                # import ipdb; ipdb.set_trace()
                 # We skip "Content-Type" and "Content-Transfer-Encoding" because we decode text payload before importing.
-                # We skip MIME-Version because it may be another one in a encrypted sub-message we take the headers from too.
+                # We skip MIME-Version because it may be another one in an encrypted sub-message we take the headers from too.
                 if k.lower() in ("content-type", "content-transfer-encoding", "mime-version"):
                     continue
                 try:
@@ -1548,7 +1588,14 @@ class Parser:
             if subtype in (HTML, PLAIN):
                 t = o.get_payload(decode=True).strip()
                 if o.get_charsets() and o.get_charsets()[0]:
-                    t = t.decode(o.get_charsets()[0])
+                    try:
+                        t = t.decode(o.get_charsets()[0])
+                    except LookupError as e:
+                        t = t.decode(errors="replace")
+                        logger.warning(f"Replacing some invalid characters in {maintype}/{subtype}: {e}")
+                    except ValueError as e:
+                        t = t.decode(o.get_charsets()[0], errors="replace")
+                        logger.warning(f"Replacing some invalid characters in {maintype}/{subtype}: {e}")
                 self.e.message(t, alternative=subtype)
             else:
                 raise ValueError(f"Unknown subtype: {subtype}")
